@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2015 David Robillard <http://drobilla.net>
+  Copyright 2011-2017 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -110,7 +110,19 @@ typedef enum {
 	   NTriples - Line-based RDF triples (ASCII).
 	   @see <a href="http://www.w3.org/TR/rdf-testcases#ntriples">NTriples</a>
 	*/
-	SERD_NTRIPLES = 2
+	SERD_NTRIPLES = 2,
+
+	/**
+	   NQuads - Line-based RDF quads (UTF-8).
+	   @see <a href="https://www.w3.org/TR/n-quads/">NQuads</a>
+	*/
+	SERD_NQUADS = 3,
+
+	/**
+	   TRiG - Terse RDF quads (UTF-8).
+	   @see <a href="https://www.w3.org/TR/trig/">Trig</a>
+	*/
+	SERD_TRIG = 4
 } SerdSyntax;
 
 /**
@@ -184,7 +196,6 @@ typedef enum {
 	   <tt>nodeID</tt></a>
 	*/
 	SERD_BLANK = 4
-
 } SerdType;
 
 /**
@@ -317,7 +328,9 @@ serd_base64_decode(const uint8_t* str, size_t len, size_t* size);
    @{
 */
 
-static const SerdURI SERD_URI_NULL = {{0,0},{0,0},{0,0},{0,0},{0,0},{0,0}};
+static const SerdURI SERD_URI_NULL = {
+	{NULL, 0}, {NULL, 0}, {NULL, 0}, {NULL, 0}, {NULL, 0}, {NULL, 0}
+};
 
 /**
    Return the local path for `uri`, or NULL if `uri` is not a file URI.
@@ -396,7 +409,7 @@ serd_uri_serialise_relative(const SerdURI* uri,
    @{
 */
 
-static const SerdNode SERD_NODE_NULL = { 0, 0, 0, 0, SERD_NOTHING };
+static const SerdNode SERD_NODE_NULL = { NULL, 0, 0, 0, SERD_NOTHING };
 
 /**
    Make a (shallow) node from `str`.
@@ -424,7 +437,7 @@ bool
 serd_node_equals(const SerdNode* a, const SerdNode* b);
 
 /**
-   Simple wrapper for serd_node_new_uri to resolve a URI node.
+   Simple wrapper for serd_node_new_uri() to resolve a URI node.
 */
 SERD_API
 SerdNode
@@ -433,7 +446,7 @@ serd_node_new_uri_from_node(const SerdNode* uri_node,
                             SerdURI*        out);
 
 /**
-   Simple wrapper for serd_node_new_uri to resolve a URI string.
+   Simple wrapper for serd_node_new_uri() to resolve a URI string.
 */
 SERD_API
 SerdNode
@@ -461,7 +474,7 @@ serd_node_new_file_uri(const uint8_t* path,
 /**
    Create a new node by serialising `uri` into a new string.
 
-   @param uri The URI to parse and serialise.
+   @param uri The URI to serialise.
 
    @param base Base URI to resolve `uri` against (or NULL for no resolution).
 
@@ -472,6 +485,25 @@ SERD_API
 SerdNode
 serd_node_new_uri(const SerdURI* uri, const SerdURI* base, SerdURI* out);
 
+/**
+   Create a new node by serialising `uri` into a new relative URI.
+
+   @param uri The URI to serialise.
+
+   @param base Base URI to make `uri` relative to, if possible.
+
+   @param root Root URI for resolution (see serd_uri_serialise_relative()).
+
+   @param out Set to the parsing of the new URI (i.e. points only to
+   memory owned by the new returned node).
+*/
+SERD_API
+SerdNode
+serd_node_new_relative_uri(const SerdURI* uri,
+                           const SerdURI* base,
+                           const SerdURI* root,
+                           SerdURI*       out);
+
 /**
    Create a new node by serialising `d` into an xsd:decimal string.
 
