@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2016 David Robillard <http://drobilla.net>
+  Copyright 2011-2017 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -55,6 +55,9 @@ typedef struct {
 	Ref                 graph;
 	Ref                 subject;
 	Ref                 predicate;
+	Ref                 object;
+	Ref                 datatype;
+	Ref                 lang;
 	SerdStatementFlags* flags;
 } ReadContext;
 
@@ -95,6 +98,18 @@ struct SerdReaderImpl {
 #endif
 };
 
+static inline bool
+supports_fancy_literals(const SerdReader* reader)
+{
+	return reader->syntax == SERD_TURTLE || reader->syntax == SERD_TRIG;
+}
+
+static inline bool
+supports_relative_iris(const SerdReader* reader)
+{
+	return reader->syntax == SERD_TURTLE || reader->syntax == SERD_TRIG;
+}
+
 static int
 r_err(SerdReader* reader, SerdStatus st, const char* fmt, ...)
 {
@@ -141,7 +156,7 @@ eat_byte_safe(SerdReader* reader, const uint8_t byte)
 {
 	assert(peek_byte(reader) == byte);
 	switch (byte) {
-	case '\0': reader->eof = true; break;
+	case '\0': reader->eof = (byte != '\0'); break;
 	case '\n': ++reader->cur.line; reader->cur.col = 0; break;
 	default:   ++reader->cur.col;
 	}
@@ -169,12 +184,14 @@ eat_byte_check(SerdReader* reader, const uint8_t byte)
 	return eat_byte_safe(reader, byte);
 }
 
-static inline void
+static inline bool
 eat_string(SerdReader* reader, const char* str, unsigned n)
 {
+	bool bad = false;
 	for (unsigned i = 0; i < n; ++i) {
-		eat_byte_check(reader, ((const uint8_t*)str)[i]);
+		bad |= eat_byte_check(reader, ((const uint8_t*)str)[i]);
 	}
+	return bad;
 }
 
 static Ref
@@ -502,13 +519,6 @@ read_ws_star(SerdReader* reader)
 	return true;
 }
 
-static inline bool
-read_ws_plus(SerdReader* reader)
-{
-	TRY_RET(read_ws(reader));
-	return read_ws_star(reader);
-}
-
 static inline bool
 peek_delim(SerdReader* reader, const char delim)
 {
@@ -618,6 +628,11 @@ read_String(SerdReader* reader, SerdNodeFlags* flags)
 		return push_node(reader, SERD_LITERAL, "", 0);
 	}
 
+	if (!supports_fancy_literals(reader)) {
+		return r_err(reader, SERD_ERR_BAD_SYNTAX,
+		             "syntax does not support long literals\n");
+	}
+
 	eat_byte_safe(reader, q3);
 	return read_STRING_LITERAL_LONG(reader, flags, q1);
 }
@@ -781,14 +796,35 @@ read_LANGTAG(SerdReader* reader)
 	return ref;
 }
 
+typedef enum { PREFIX, GOOD, BAD} SchemeState;
+
+static inline bool
+check_scheme(SerdReader* reader, uint8_t c, SchemeState* state)
+{
+	if (!supports_relative_iris(reader) && *state == PREFIX) {
+		if (c == ':') {
+			*state = GOOD;
+		} else if (!isalpha(c)) {
+			*state = BAD;
+			return r_err(reader, SERD_ERR_BAD_SYNTAX,
+			             "syntax does not support relative IRIs\n");
+		}
+	}
+	return true;
+}
+
 static Ref
 read_IRIREF(SerdReader* reader)
 {
 	TRY_RET(eat_byte_check(reader, '<'));
-	Ref      ref = push_node(reader, SERD_URI, "", 0);
-	uint32_t code;
+	Ref         ref    = push_node(reader, SERD_URI, "", 0);
+	SchemeState scheme = PREFIX;
+	uint32_t    code;
 	while (true) {
 		const uint8_t c = peek_byte(reader);
+		if (!check_scheme(reader, c, &scheme)) {
+			return pop_node(reader, ref);
+		}
 		switch (c) {
 		case '"': case '<': case '^': case '`': case '{': case '|': case '}':
 			r_err(reader, SERD_ERR_BAD_SYNTAX,
@@ -833,17 +869,13 @@ read_IRIREF(SerdReader* reader)
 static bool
 read_PrefixedName(SerdReader* reader, Ref dest, bool read_prefix, bool* ate_dot)
 {
-	if (read_prefix) {
-		if (read_PN_PREFIX(reader, dest) > SERD_FAILURE) {
-			return false;
-		}
-	}
-
-	if (eat_byte_check(reader, ':') != ':') {
+	if (read_prefix && read_PN_PREFIX(reader, dest) > SERD_FAILURE) {
+		return false;
+	} else if (peek_byte(reader) != ':') {
 		return false;
 	}
 
-	push_byte(reader, dest, ':');
+	push_byte(reader, dest, eat_byte_safe(reader, ':'));
 	return read_PN_LOCAL(reader, dest, ate_dot) <= SERD_FAILURE;
 }
 
@@ -930,14 +962,11 @@ read_iri(SerdReader* reader, Ref* dest, bool* ate_dot)
 	switch (peek_byte(reader)) {
 	case '<':
 		*dest = read_IRIREF(reader);
-		break;
+		return true;
 	default:
 		*dest = push_node(reader, SERD_CURIE, "", 0);
-		if (!read_PrefixedName(reader, *dest, true, ate_dot)) {
-			*dest = pop_node(reader, *dest);
-		}
+		return read_PrefixedName(reader, *dest, true, ate_dot);
 	}
-	return *dest != 0;
 }
 
 static bool
@@ -963,8 +992,8 @@ read_literal(SerdReader* reader, Ref* dest,
 	*dest = str;
 	return true;
 except:
-	pop_node(reader, *datatype);
-	pop_node(reader, *lang);
+	*datatype = pop_node(reader, *datatype);
+	*lang     = pop_node(reader, *lang);
 	pop_node(reader, str);
 	return false;
 }
@@ -1141,10 +1170,11 @@ read_anon(SerdReader* reader, ReadContext ctx, bool subject, Ref* dest)
 	return (eat_byte_check(reader, ']') == ']');
 }
 
-// Recurses, calling statement_sink for every statement encountered.
-// Leaves stack in original calling state (i.e. pops everything it pushes).
+/* If emit is true: recurses, calling statement_sink for every statement
+   encountered, and leaves stack in original calling state (i.e. pops
+   everything it pushes). */
 static bool
-read_object(SerdReader* reader, ReadContext ctx, bool* ate_dot)
+read_object(SerdReader* reader, ReadContext* ctx, bool emit, bool* ate_dot)
 {
 	static const char* const XSD_BOOLEAN     = NS_XSD "boolean";
 	static const size_t      XSD_BOOLEAN_LEN = 40;
@@ -1154,24 +1184,31 @@ read_object(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 #endif
 
 	bool          ret      = false;
-	bool          emit     = (ctx.subject != 0);
+	bool          simple   = (ctx->subject != 0);
 	SerdNode*     node     = NULL;
 	Ref           o        = 0;
 	Ref           datatype = 0;
 	Ref           lang     = 0;
 	uint32_t      flags    = 0;
 	const uint8_t c        = peek_byte(reader);
+	if (!supports_fancy_literals(reader)) {
+		switch (c) {
+		case '"': case ':': case '<': case '_': break;
+		default: return r_err(reader, SERD_ERR_BAD_SYNTAX,
+		                      "expected: ':', '<', or '_'\n");
+		}
+	}
 	switch (c) {
 	case '\0':
 	case ')':
 		return false;
 	case '[':
-		emit = false;
-		TRY_THROW(ret = read_anon(reader, ctx, false, &o));
+		simple = false;
+		TRY_THROW(ret = read_anon(reader, *ctx, false, &o));
 		break;
 	case '(':
-		emit = false;
-		TRY_THROW(ret = read_collection(reader, ctx, &o));
+		simple = false;
+		TRY_THROW(ret = read_collection(reader, *ctx, &o));
 		break;
 	case '_':
 		TRY_THROW(ret = (o = read_BLANK_NODE_LABEL(reader, ate_dot)));
@@ -1206,9 +1243,17 @@ read_object(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 		}
 	}
 
-	if (ret && emit) {
+	if (simple && o) {
 		deref(reader, o)->flags = flags;
-		ret = emit_statement(reader, ctx, o, datatype, lang);
+	}
+
+	if (ret && emit && simple) {
+		ret = emit_statement(reader, *ctx, o, datatype, lang);
+	} else if (ret && !emit) {
+		ctx->object   = o;
+		ctx->datatype = datatype;
+		ctx->lang     = lang;
+		return true;
 	}
 
 except:
@@ -1224,9 +1269,9 @@ except:
 static bool
 read_objectList(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 {
-	TRY_RET(read_object(reader, ctx, ate_dot));
+	TRY_RET(read_object(reader, &ctx, true, ate_dot));
 	while (!*ate_dot && eat_delim(reader, ',')) {
-		TRY_RET(read_object(reader, ctx, ate_dot));
+		TRY_RET(read_object(reader, &ctx, true, ate_dot));
 	}
 	return true;
 }
@@ -1251,7 +1296,7 @@ read_predicateObjectList(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 			switch (c = peek_byte(reader)) {
 			case 0:
 				return false;
-			case '.': case ']':
+			case '.': case ']': case '}':
 				return true;
 			case ';':
 				eat_byte_safe(reader, c);
@@ -1311,7 +1356,7 @@ read_collection(SerdReader* reader, ReadContext ctx, Ref* dest)
 		// _:node rdf:first object
 		ctx.predicate = reader->rdf_first;
 		bool ate_dot = false;
-		if (!read_object(reader, ctx, &ate_dot) || ate_dot) {
+		if (!read_object(reader, &ctx, true, &ate_dot) || ate_dot) {
 			return end_collection(reader, ctx, n1, n2, false);
 		}
 
@@ -1340,80 +1385,98 @@ read_collection(SerdReader* reader, ReadContext ctx, Ref* dest)
 }
 
 static Ref
-read_subject(SerdReader* reader, ReadContext ctx, bool* nested)
+read_subject(SerdReader* reader, ReadContext ctx, Ref* dest, char* s_type)
+{
+	bool ate_dot = false;
+	switch ((*s_type = peek_byte(reader))) {
+	case '[':
+		read_anon(reader, ctx, true, dest);
+		break;
+	case '(':
+		read_collection(reader, ctx, dest);
+		break;
+	case '_':
+		*dest = read_BLANK_NODE_LABEL(reader, &ate_dot);
+		break;
+	default:
+		TRY_RET(read_iri(reader, dest, &ate_dot));
+	}
+	return ate_dot ? pop_node(reader, *dest) : *dest;
+}
+
+static Ref
+read_labelOrSubject(SerdReader* reader, ReadContext ctx)
 {
 	Ref  subject = 0;
 	bool ate_dot = false;
 	switch (peek_byte(reader)) {
 	case '[':
-		*nested = true;
-		read_anon(reader, ctx, true, &subject);
-		break;
-	case '(':
-		*nested = true;
-		read_collection(reader, ctx, &subject);
-		break;
+		eat_byte_safe(reader, '[');
+		read_ws_star(reader);
+		TRY_RET(eat_byte_check(reader, ']'));
+		return blank_id(reader);
 	case '_':
-		*nested = false;
-		subject = read_BLANK_NODE_LABEL(reader, &ate_dot);
-		break;
+		return read_BLANK_NODE_LABEL(reader, &ate_dot);
 	default:
 		read_iri(reader, &subject, &ate_dot);
 	}
-	return ate_dot ? pop_node(reader, subject) : subject;
+	return subject;
 }
 
 static bool
 read_triples(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 {
-	bool      nested  = false;
-	const Ref subject = read_subject(reader, ctx, &nested);
-	bool      ret     = false;
-	if (subject) {
-		ctx.subject = subject;
-		if (nested) {
-			read_ws_star(reader);
-			ret = true;
-			if (peek_byte(reader) != '.') {
-				ret = read_predicateObjectList(reader, ctx, ate_dot);
-			}
-		} else {
-			TRY_RET(read_ws_plus(reader));
-			ret = read_predicateObjectList(reader, ctx, ate_dot);
+	bool ret = false;
+	if (ctx.subject) {
+		read_ws_star(reader);
+		switch (peek_byte(reader)) {
+		case '.':
+			*ate_dot = eat_byte_safe(reader, '.');
+			return false;
+		case '}':
+			return false;
 		}
-		pop_node(reader, subject);
-	} else {
-		ret = false;
+		ret = read_predicateObjectList(reader, ctx, ate_dot);
 	}
 	ctx.subject = ctx.predicate = 0;
 	return ret;
 }
 
 static bool
-read_base(SerdReader* reader)
+read_base(SerdReader* reader, bool sparql, bool token)
 {
-	// `@' is already eaten in read_directive
-	eat_string(reader, "base", 4);
-	TRY_RET(read_ws_plus(reader));
+	if (token) {
+		TRY_RET(eat_string(reader, "base", 4));
+	}
+
 	Ref uri;
+	read_ws_star(reader);
 	TRY_RET(uri = read_IRIREF(reader));
 	if (reader->base_sink) {
 		reader->base_sink(reader->handle, deref(reader, uri));
 	}
 	pop_node(reader, uri);
+
+	read_ws_star(reader);
+	if (!sparql) {
+		return eat_byte_check(reader, '.');
+	} else if (peek_byte(reader) == '.') {
+		return r_err(reader, SERD_ERR_BAD_SYNTAX,
+		             "full stop after SPARQL BASE\n");
+	}
 	return true;
 }
 
 static bool
-read_prefixID(SerdReader* reader)
+read_prefixID(SerdReader* reader, bool sparql, bool token)
 {
-	bool ret = true;
-	// `@' is already eaten in read_directive
-	eat_string(reader, "prefix", 6);
-	TRY_RET(read_ws_plus(reader));
-
-	Ref name = push_node(reader, SERD_LITERAL, "", 0);
+	if (token) {
+		TRY_RET(eat_string(reader, "prefix", 6));
+	}
 
+	read_ws_star(reader);
+	bool ret  = true;
+	Ref  name = push_node(reader, SERD_LITERAL, "", 0);
 	if (read_PN_PREFIX(reader, name) > SERD_FAILURE) {
 		return pop_node(reader, name);
 	}
@@ -1436,27 +1499,87 @@ read_prefixID(SerdReader* reader)
 	}
 	pop_node(reader, uri);
 	pop_node(reader, name);
+	if (!sparql) {
+		read_ws_star(reader);
+		return eat_byte_check(reader, '.');
+	}
 	return ret;
 }
 
 static bool
 read_directive(SerdReader* reader)
 {
-	eat_byte_safe(reader, '@');
-	switch (peek_byte(reader)) {
-	case 'b': return read_base(reader);
-	case 'p': return read_prefixID(reader);
-	default:  return r_err(reader, SERD_ERR_BAD_SYNTAX, "invalid directive\n");
+	const bool sparql = peek_byte(reader) != '@';
+	if (!sparql) {
+		eat_byte_safe(reader, '@');
+		switch (peek_byte(reader)) {
+		case 'B': case 'P':
+			return r_err(reader, SERD_ERR_BAD_SYNTAX,
+			             "uppercase directive\n");
+		}
 	}
+
+	switch (peek_byte(reader)) {
+	case 'B': case 'b': return read_base(reader, sparql, true);
+	case 'P': case 'p': return read_prefixID(reader, sparql, true);
+	default:
+		return r_err(reader, SERD_ERR_BAD_SYNTAX, "invalid directive\n");
+	}
+
+	return true;
+}
+
+static bool
+read_wrappedGraph(SerdReader* reader, ReadContext* ctx)
+{
+	bool ate_dot = false;
+	char s_type  = 0;
+	TRY_RET(eat_byte_check(reader, '{'));
+	read_ws_star(reader);
+	while (peek_byte(reader) != '}') {
+		ctx->subject = 0;
+		Ref subj = read_subject(reader, *ctx, &ctx->subject, &s_type);
+		if (!subj ||
+		    (!read_triples(reader, *ctx, &ate_dot) && s_type != '[')) {
+			return false;
+		}
+		pop_node(reader, subj);
+		read_ws_star(reader);
+		if (peek_byte(reader) == '.') {
+			eat_byte_safe(reader, '.');
+		}
+		read_ws_star(reader);
+	}
+	return eat_byte_check(reader, '}');
+}
+
+static int
+tokcmp(SerdReader* reader, Ref ref, const char* tok, size_t n)
+{
+	SerdNode* node = deref(reader, ref);
+	if (!node || node->n_bytes != n) {
+		return -1;
+	}
+	const char* s1 = (const char*)node->buf;
+	const char* s2 = tok;
+	for (; n > 0 && *s2; s1++, s2++, --n) {
+		if (toupper(*s1) != toupper(*s2)) {
+			return ((*(uint8_t*)s1 < *(uint8_t*)s2) ? -1 : +1);
+		}
+	}
+    return 0;
 }
 
 static bool
 read_statement(SerdReader* reader)
 {
-	SerdStatementFlags flags = 0;
-	ReadContext ctx = { 0, 0, 0, &flags };
+	SerdStatementFlags flags   = 0;
+	ReadContext        ctx     = { 0, 0, 0, 0, 0, 0, &flags };
+	Ref                subj    = 0;
+	bool               ate_dot = false;
+	char               s_type  = false;
+	bool               ret     = true;
 	read_ws_star(reader);
-	bool ate_dot = false;
 	switch (peek_byte(reader)) {
 	case '\0':
 		reader->eof = true;
@@ -1464,20 +1587,47 @@ read_statement(SerdReader* reader)
 	case '@':
 		TRY_RET(read_directive(reader));
 		read_ws_star(reader);
-		return (eat_byte_check(reader, '.') == '.');
-	default:
-		if (!read_triples(reader, ctx, &ate_dot)) {
-			return false;
-		} else if (ate_dot) {
-			return true;
-		} else {
+		break;
+	case '{':
+		if (reader->syntax == SERD_TRIG) {
+			TRY_RET(read_wrappedGraph(reader, &ctx));
 			read_ws_star(reader);
-			return (eat_byte_check(reader, '.') == '.');
+		} else {
+			return r_err(reader, SERD_ERR_BAD_SYNTAX, "graph in Turtle\n");
 		}
 		break;
+	default:
+		subj = read_subject(reader, ctx, &ctx.subject, &s_type);
+		if (!tokcmp(reader, ctx.subject, "base", 4)) {
+			ret = read_base(reader, true, false);
+		} else if (!tokcmp(reader, ctx.subject, "prefix", 6)) {
+			ret = read_prefixID(reader, true, false);
+		} else if (!tokcmp(reader, ctx.subject, "graph", 5)) {
+			read_ws_star(reader);
+			TRY_RET((ctx.graph = read_labelOrSubject(reader, ctx)));
+			read_ws_star(reader);
+			TRY_RET(read_wrappedGraph(reader, &ctx));
+			read_ws_star(reader);
+		} else if (read_ws_star(reader) && peek_byte(reader) == '{') {
+			if (s_type == '(' || (s_type == '[' && !*ctx.flags)) {
+				return false;  // invalid graph with complex label
+			}
+			ctx.graph   = subj;
+			ctx.subject = subj = 0;
+			TRY_RET(read_wrappedGraph(reader, &ctx));
+			read_ws_star(reader);
+		} else if (!subj) {
+			ret = r_err(reader, SERD_ERR_BAD_SYNTAX, "bad subject\n");
+		} else if (!read_triples(reader, ctx, &ate_dot)) {
+			ret = (s_type == '[');
+		} else if (!ate_dot) {
+			read_ws_star(reader);
+			ret = (eat_byte_check(reader, '.') == '.');
+		}
+		pop_node(reader, subj);
+		break;
 	}
-	read_ws_star(reader); // remove?
-	return true;
+	return ret;
 }
 
 static bool
@@ -1489,6 +1639,76 @@ read_turtleDoc(SerdReader* reader)
 	return !reader->error;
 }
 
+static bool
+read_trigDoc(SerdReader* reader)
+{
+	while (!reader->eof) {
+		TRY_RET(read_statement(reader));
+	}
+	return !reader->error;
+}
+
+static bool
+read_nquadsDoc(SerdReader* reader)
+{
+	while (!reader->eof) {
+		SerdStatementFlags flags   = 0;
+		ReadContext        ctx     = { 0, 0, 0, 0, 0, 0, &flags };
+		bool               ate_dot = false;
+		char               s_type  = false;
+		read_ws_star(reader);
+		if (peek_byte(reader) == '\0') {
+			reader->eof = true;
+			break;
+		}
+
+		// subject predicate object
+		if (!(ctx.subject = read_subject(reader, ctx, &ctx.subject, &s_type)) ||
+		    !read_ws_star(reader) ||
+		    !(ctx.predicate = read_IRIREF(reader)) ||
+		    !read_ws_star(reader) ||
+		    !read_object(reader, &ctx, false, &ate_dot)) {
+			return false;
+		}
+
+		if (!ate_dot) {  // graphLabel?
+			TRY_RET(read_ws_star(reader));
+			switch (peek_byte(reader)) {
+			case '.':
+				break;
+			case '_':
+				ctx.graph = read_BLANK_NODE_LABEL(reader, &ate_dot);
+				break;
+			default:
+				if (!(ctx.graph = read_IRIREF(reader))) {
+					return false;
+				}
+			}
+
+			// Terminating '.'
+			TRY_RET(read_ws_star(reader));
+			eat_byte_check(reader, '.');
+		}
+
+		TRY_RET(emit_statement(reader, ctx, ctx.object, ctx.datatype, ctx.lang));
+		pop_node(reader, ctx.graph);
+		pop_node(reader, ctx.lang);
+		pop_node(reader, ctx.datatype);
+		pop_node(reader, ctx.object);
+	}
+	return !reader->error;
+}
+
+static bool
+read_doc(SerdReader* reader)
+{
+	switch (reader->syntax) {
+	case SERD_NQUADS: return read_nquadsDoc(reader);
+	case SERD_TRIG:   return read_trigDoc(reader);
+	default:          return read_turtleDoc(reader);
+	}
+}
+
 SERD_API
 SerdReader*
 serd_reader_new(SerdSyntax        syntax,
@@ -1686,7 +1906,7 @@ serd_reader_read_file_handle(SerdReader* me, FILE* file, const uint8_t* name)
 {
 	SerdStatus st = serd_reader_start_stream(me, file, name, true);
 	if (!st) {
-		st = read_turtleDoc(me) ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
+		st = read_doc(me) ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
 		serd_reader_end_stream(me);
 	}
 	return st;
@@ -1706,7 +1926,7 @@ serd_reader_read_string(SerdReader* me, const uint8_t* utf8)
 	me->eof       = false;
 
 	skip_bom(me);
-	const bool ret = read_turtleDoc(me);
+	const bool ret = read_doc(me);
 
 	me->read_buf = NULL;
 	return ret ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
