@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2015 David Robillard <http://drobilla.net>
+  Copyright 2011-2017 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -47,7 +47,9 @@ typedef enum {
 	SEP_ANON_END,    ///< End of anonymous node (']')
 	SEP_LIST_BEGIN,  ///< Start of list ('(')
 	SEP_LIST_SEP,    ///< List separator (whitespace)
-	SEP_LIST_END     ///< End of list (')')
+	SEP_LIST_END,    ///< End of list (')')
+	SEP_GRAPH_BEGIN, ///< Start of graph ('{')
+	SEP_GRAPH_END    ///< End of graph ('}')
 } Sep;
 
 typedef struct {
@@ -70,6 +72,8 @@ static const SepRule rules[] = {
 	{ "(",      1, 0, 0, 0 },
 	{ NULL,     1, 0, 1, 0 },
 	{ ")",      1, 1, 0, 0 },
+	{ " {",     2, 0, 1, 1 },
+	{ " }",     2, 0, 1, 1 },
 	{ "\n",     1, 0, 1, 0 }
 };
 
@@ -101,6 +105,12 @@ typedef enum {
 	WRITE_LONG_STRING
 } TextContext;
 
+static bool
+supports_abbrev(const SerdWriter* writer)
+{
+	return writer->syntax == SERD_TURTLE || writer->syntax == SERD_TRIG;
+}
+
 static void
 w_err(SerdWriter* writer, SerdStatus st, const char* fmt, ...)
 {
@@ -252,7 +262,7 @@ write_uri(SerdWriter* writer, const uint8_t* utf8, size_t n_bytes)
 
 		// Bulk write all characters up to this special one
 		len += sink(&utf8[i], j - i, writer);
- 		if ((i = j) == n_bytes) {
+		if ((i = j) == n_bytes) {
 			break;  // Reached end
 		}
 
@@ -299,7 +309,7 @@ write_lname(SerdWriter* writer, const uint8_t* utf8, size_t n_bytes)
 
 		// Bulk write all characters up to this special one
 		len += sink(&utf8[i], j - i, writer);
- 		if ((i = j) == n_bytes) {
+		if ((i = j) == n_bytes) {
 			break;  // Reached end
 		}
 
@@ -356,7 +366,7 @@ write_text(SerdWriter* writer, TextContext ctx,
 			case '"':  len += sink("\\\"", 2, writer); continue;
 			default: break;
 			}
-			if (writer->syntax != SERD_NTRIPLES) {
+			if (writer->syntax == SERD_TURTLE) {
 				switch (in) {
 				case '\b': len += sink("\\b", 2, writer); continue;
 				case '\f': len += sink("\\f", 2, writer); continue;
@@ -411,33 +421,38 @@ write_sep(SerdWriter* writer, const Sep sep)
 }
 
 static SerdStatus
-reset_context(SerdWriter* writer, bool del)
+reset_context(SerdWriter* writer, bool graph)
 {
-	if (del) {
-		serd_node_free(&writer->context.graph);
-		serd_node_free(&writer->context.subject);
-		serd_node_free(&writer->context.predicate);
-		writer->context = WRITE_CONTEXT_NULL;
-	} else {
-		writer->context.graph.type     = SERD_NOTHING;
-		writer->context.subject.type   = SERD_NOTHING;
-		writer->context.predicate.type = SERD_NOTHING;
+	if (graph) {
+		writer->context.graph.type = SERD_NOTHING;
 	}
-	writer->empty = false;
+	writer->context.subject.type   = SERD_NOTHING;
+	writer->context.predicate.type = SERD_NOTHING;
+	writer->empty                  = false;
 	return SERD_SUCCESS;
 }
 
+static SerdStatus
+free_context(SerdWriter* writer)
+{
+	serd_node_free(&writer->context.graph);
+	serd_node_free(&writer->context.subject);
+	serd_node_free(&writer->context.predicate);
+	return reset_context(writer, true);
+}
+
 typedef enum {
 	FIELD_NONE,
 	FIELD_SUBJECT,
 	FIELD_PREDICATE,
-	FIELD_OBJECT
+	FIELD_OBJECT,
+	FIELD_GRAPH
 } Field;
 
 static bool
 is_inline_start(const SerdWriter* writer, Field field, SerdStatementFlags flags)
 {
-	return (writer->syntax != SERD_NTRIPLES &&
+	return (supports_abbrev(writer) &&
 	        ((field == FIELD_SUBJECT && (flags & SERD_ANON_S_BEGIN)) ||
 	         (field == FIELD_OBJECT &&  (flags & SERD_ANON_O_BEGIN))));
 }
@@ -459,19 +474,19 @@ write_node(SerdWriter*        writer,
 		if (is_inline_start(writer, field, flags)) {
 			++writer->indent;
 			write_sep(writer, SEP_ANON_BEGIN);
-		} else if (writer->syntax != SERD_NTRIPLES
+		} else if (supports_abbrev(writer)
 		           && (field == FIELD_SUBJECT && (flags & SERD_LIST_S_BEGIN))) {
 			assert(writer->list_depth == 0);
 			copy_node(&writer->list_subj, node);
 			++writer->list_depth;
 			++writer->indent;
 			write_sep(writer, SEP_LIST_BEGIN);
-		} else if (writer->syntax != SERD_NTRIPLES
+		} else if (supports_abbrev(writer)
 		           && (field == FIELD_OBJECT && (flags & SERD_LIST_O_BEGIN))) {
 			++writer->indent;
 			++writer->list_depth;
 			write_sep(writer, SEP_LIST_BEGIN);
-		} else if (writer->syntax != SERD_NTRIPLES
+		} else if (supports_abbrev(writer)
 		           && ((field == FIELD_SUBJECT && (flags & SERD_EMPTY_S))
 		               || (field == FIELD_OBJECT && (flags & SERD_EMPTY_O)))) {
 			sink("[]", 2, writer);
@@ -491,6 +506,7 @@ write_node(SerdWriter*        writer,
 	case SERD_CURIE:
 		switch (writer->syntax) {
 		case SERD_NTRIPLES:
+		case SERD_NQUADS:
 			if (serd_env_expand(writer->env, node, &uri_prefix, &suffix)) {
 				w_err(writer, SERD_ERR_BAD_CURIE,
 				      "undefined namespace prefix `%s'\n", node->buf);
@@ -502,6 +518,7 @@ write_node(SerdWriter*        writer,
 			sink(">", 1, writer);
 			break;
 		case SERD_TURTLE:
+		case SERD_TRIG:
 			if (is_inline_start(writer, field, flags)) {
 				++writer->indent;
 				write_sep(writer, SEP_ANON_BEGIN);
@@ -515,14 +532,15 @@ write_node(SerdWriter*        writer,
 		}
 		break;
 	case SERD_LITERAL:
-		if (writer->syntax == SERD_TURTLE && datatype && datatype->buf) {
+		if (supports_abbrev(writer) && datatype && datatype->buf) {
 			const char* type_uri = (const char*)datatype->buf;
 			if (!strncmp(type_uri, NS_XSD, sizeof(NS_XSD) - 1) && (
 				    !strcmp(type_uri + sizeof(NS_XSD) - 1, "boolean") ||
 				    !strcmp(type_uri + sizeof(NS_XSD) - 1, "integer"))) {
 				sink(node->buf, node->n_bytes, writer);
 				break;
-			} else if (!strcmp(type_uri + sizeof(NS_XSD) - 1, "decimal") &&
+			} else if (!strncmp(type_uri, NS_XSD, sizeof(NS_XSD) - 1) &&
+			           !strcmp(type_uri + sizeof(NS_XSD) - 1, "decimal") &&
 			           strchr((const char*)node->buf, '.') &&
 			           node->buf[node->n_bytes - 1] != '.') {
 				/* xsd:decimal literals without trailing digits, e.g. "5.", can
@@ -533,7 +551,7 @@ write_node(SerdWriter*        writer,
 				break;
 			}
 		}
-		if (writer->syntax != SERD_NTRIPLES
+		if (supports_abbrev(writer)
 		    && (node->flags & (SERD_HAS_NEWLINE|SERD_HAS_QUOTE))) {
 			sink("\"\"\"", 3, writer);
 			write_text(writer, WRITE_LONG_STRING, node->buf, node->n_bytes);
@@ -558,11 +576,11 @@ write_node(SerdWriter*        writer,
 			sink("== ", 3, writer);
 		}
 		has_scheme = serd_uri_string_has_scheme(node->buf);
-		if (field == FIELD_PREDICATE && (writer->syntax == SERD_TURTLE)
+		if (field == FIELD_PREDICATE && supports_abbrev(writer)
 		    && !strcmp((const char*)node->buf, NS_RDF "type")) {
 			sink("a", 1, writer);
 			break;
-		} else if ((writer->syntax == SERD_TURTLE)
+		} else if (supports_abbrev(writer)
 		           && !strcmp((const char*)node->buf, NS_RDF "nil")) {
 			sink("()", 2, writer);
 			break;
@@ -582,7 +600,8 @@ write_node(SerdWriter*        writer,
 			bool rooted = uri_is_under(&writer->base_uri, &writer->root_uri);
 			SerdURI* root = rooted ? &writer->root_uri : & writer->base_uri;
 			if (!uri_is_under(&abs_uri, root) ||
-			    writer->syntax == SERD_NTRIPLES) {
+			    writer->syntax == SERD_NTRIPLES ||
+			    writer->syntax == SERD_NQUADS) {
 				serd_uri_serialise(&abs_uri, uri_sink, writer);
 			} else {
 				serd_uri_serialise_relative(
@@ -660,22 +679,47 @@ serd_writer_write_statement(SerdWriter*        writer,
 
 	switch (writer->syntax) {
 	case SERD_NTRIPLES:
+	case SERD_NQUADS:
 		TRY(write_node(writer, subject, NULL, NULL, FIELD_SUBJECT, flags));
 		sink(" ", 1, writer);
 		TRY(write_node(writer, predicate, NULL, NULL, FIELD_PREDICATE, flags));
 		sink(" ", 1, writer);
 		TRY(write_node(writer, object, datatype, lang, FIELD_OBJECT, flags));
+		if (writer->syntax == SERD_NQUADS && graph) {
+			sink(" ", 1, writer);
+			TRY(write_node(writer, graph, datatype, lang, FIELD_GRAPH, flags));
+		}
 		sink(" .\n", 3, writer);
 		return SERD_SUCCESS;
 	default:
 		break;
 	}
 
+	if ((graph && !serd_node_equals(graph, &writer->context.graph)) ||
+	    (!graph && writer->context.graph.type)) {
+		writer->indent = 0;
+		if (writer->context.subject.type) {
+			write_sep(writer, SEP_END_S);
+		}
+		if (writer->context.graph.type) {
+			write_sep(writer, SEP_GRAPH_END);
+		}
+
+		reset_context(writer, true);
+		if (graph) {
+			TRY(write_node(writer, graph, datatype, lang, FIELD_GRAPH, flags));
+			++writer->indent;
+			write_sep(writer, SEP_GRAPH_BEGIN);
+			copy_node(&writer->context.graph, graph);
+		}
+	}
+
 	if ((flags & SERD_LIST_CONT)) {
 		if (write_list_obj(writer, flags, predicate, object, datatype, lang)) {
 			// Reached end of list
 			if (--writer->list_depth == 0 && writer->list_subj.type) {
-				reset_context(writer, true);
+				reset_context(writer, false);
+				serd_node_free(&writer->context.subject);
 				writer->context.subject = writer->list_subj;
 				writer->list_subj       = SERD_NODE_NULL;
 			}
@@ -719,7 +763,7 @@ serd_writer_write_statement(SerdWriter*        writer,
 			++writer->indent;
 		}
 
-		reset_context(writer, true);
+		reset_context(writer, false);
 		copy_node(&writer->context.subject, subject);
 
 		if (!(flags & SERD_LIST_S_BEGIN)) {
@@ -753,7 +797,7 @@ SerdStatus
 serd_writer_end_anon(SerdWriter*     writer,
                      const SerdNode* node)
 {
-	if (writer->syntax == SERD_NTRIPLES) {
+	if (writer->syntax == SERD_NTRIPLES || writer->syntax == SERD_NQUADS) {
 		return SERD_SUCCESS;
 	}
 	if (serd_stack_is_empty(&writer->anon_stack) || writer->indent == 0) {
@@ -763,7 +807,7 @@ serd_writer_end_anon(SerdWriter*     writer,
 	}
 	--writer->indent;
 	write_sep(writer, SEP_ANON_END);
-	reset_context(writer, true);
+	free_context(writer);
 	writer->context = *anon_stack_top(writer);
 	serd_stack_pop(&writer->anon_stack, sizeof(WriteContext));
 	const bool is_subject = serd_node_equals(node, &writer->context.subject);
@@ -781,11 +825,14 @@ serd_writer_finish(SerdWriter* writer)
 	if (writer->context.subject.type) {
 		sink(" .\n", 3, writer);
 	}
+	if (writer->context.graph.type) {
+		sink("}\n", 2, writer);
+	}
 	if (writer->style & SERD_STYLE_BULK) {
 		serd_bulk_sink_flush(&writer->bulk_sink);
 	}
 	writer->indent = 0;
-	return reset_context(writer, true);
+	return free_context(writer);
 }
 
 SERD_API
@@ -798,7 +845,7 @@ serd_writer_new(SerdSyntax     syntax,
                 void*          stream)
 {
 	const WriteContext context = WRITE_CONTEXT_NULL;
-	SerdWriter*        writer  = (SerdWriter*)malloc(sizeof(SerdWriter));
+	SerdWriter*        writer  = (SerdWriter*)calloc(1, sizeof(SerdWriter));
 	writer->syntax       = syntax;
 	writer->style        = style;
 	writer->env          = env;
@@ -808,15 +855,8 @@ serd_writer_new(SerdSyntax     syntax,
 	writer->anon_stack   = serd_stack_new(sizeof(WriteContext));
 	writer->sink         = ssink;
 	writer->stream       = stream;
-	writer->error_sink   = NULL;
-	writer->error_handle = NULL;
 	writer->context      = context;
 	writer->list_subj    = SERD_NODE_NULL;
-	writer->list_depth   = 0;
-	writer->bprefix      = NULL;
-	writer->bprefix_len  = 0;
-	writer->indent       = 0;
-	writer->last_sep     = SEP_NONE;
 	writer->empty        = true;
 	if (style & SERD_STYLE_BULK) {
 		writer->bulk_sink = serd_bulk_sink_new(ssink, stream, SERD_PAGE_SIZE);
@@ -857,17 +897,17 @@ serd_writer_set_base_uri(SerdWriter*     writer,
 	if (!serd_env_set_base_uri(writer->env, uri)) {
 		serd_env_get_base_uri(writer->env, &writer->base_uri);
 
-		if (writer->syntax != SERD_NTRIPLES) {
+		if (writer->syntax == SERD_TURTLE || writer->syntax == SERD_TRIG) {
 			if (writer->context.graph.type || writer->context.subject.type) {
 				sink(" .\n\n", 4, writer);
-				reset_context(writer, false);
+				reset_context(writer, true);
 			}
 			sink("@base <", 7, writer);
 			sink(uri->buf, uri->n_bytes, writer);
 			sink("> .\n", 4, writer);
 		}
 		writer->indent = 0;
-		return reset_context(writer, false);
+		return reset_context(writer, true);
 	}
 	return SERD_ERR_UNKNOWN;
 }
@@ -895,10 +935,10 @@ serd_writer_set_prefix(SerdWriter*     writer,
                        const SerdNode* uri)
 {
 	if (!serd_env_set_prefix(writer->env, name, uri)) {
-		if (writer->syntax != SERD_NTRIPLES) {
+		if (writer->syntax == SERD_TURTLE || writer->syntax == SERD_TRIG) {
 			if (writer->context.graph.type || writer->context.subject.type) {
 				sink(" .\n\n", 4, writer);
-				reset_context(writer, false);
+				reset_context(writer, true);
 			}
 			sink("@prefix ", 8, writer);
 			sink(name->buf, name->n_bytes, writer);
@@ -907,7 +947,7 @@ serd_writer_set_prefix(SerdWriter*     writer,
 			sink("> .\n", 4, writer);
 		}
 		writer->indent = 0;
-		return reset_context(writer, false);
+		return reset_context(writer, true);
 	}
 	return SERD_ERR_UNKNOWN;
 }
