@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2014 David Robillard <http://drobilla.net>
+  Copyright 2011-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -332,6 +332,23 @@ main(void)
 	free(out_path);
 	serd_node_free(&file_node);
 
+	// File URI with space and no escaping
+	path_str  = USTR("a/relative path");
+	file_node = serd_node_new_file_uri(path_str, 0, 0, false);
+	out_path  = serd_file_uri_parse(file_node.buf, &hostname);
+	if (strcmp((const char*)file_node.buf, "a/relative path")) {
+		return failure("Bad URI %s\n", file_node.buf);
+	} else if (hostname) {
+		return failure("hostname `%s' shouldn't exist\n", hostname);
+	} else if (strcmp((const char*)path_str, (const char*)out_path)) {
+		return failure("path=>URI=>path failure %s => %s => %s\n",
+		               path_str, file_node.buf, out_path);
+	}
+	free(hostname);
+	free(out_path);
+	serd_node_free(&file_node);
+
+	// File URI with space and escaping
 	path_str  = USTR("a/relative path");
 	file_node = serd_node_new_file_uri(path_str, 0, 0, false);
 	out_path  = serd_file_uri_parse(file_node.buf, &hostname);
@@ -415,10 +432,23 @@ main(void)
 	    nil2.type != SERD_URI || strcmp((const char*)nil2.buf, (const char*)base.buf)) {
 		return failure("URI %s != base %s\n", nil.buf, base.buf);
 	}
-	serd_node_free(&base);
 	serd_node_free(&nil);
 	serd_node_free(&nil2);
 
+	// Test serd_node_new_relative_uri
+	SerdNode abs = serd_node_from_string(SERD_URI, USTR("http://example.org/foo/bar"));
+	SerdURI  abs_uri;
+	serd_uri_parse(abs.buf, &abs_uri);
+
+	SerdURI  rel_uri;
+	SerdNode rel = serd_node_new_relative_uri(&abs_uri, &base_uri, NULL, &rel_uri);
+	if (strcmp((const char*)rel.buf, "/foo/bar")) {
+		return failure("Bad relative URI %s (expected '/foo/bar')\n", rel.buf);
+	}
+
+	serd_node_free(&rel);
+	serd_node_free(&base);
+
 	// Test SerdEnv
 
 	SerdNode u   = serd_node_from_string(SERD_URI, USTR("http://example.org/foo"));
@@ -602,10 +632,7 @@ main(void)
 	// Rewind and test reader
 	fseek(fd, 0, SEEK_SET);
 
-	ReaderTest* rt   = (ReaderTest*)malloc(sizeof(ReaderTest));
-	rt->n_statements = 0;
-	rt->graph        = NULL;
-
+	ReaderTest* rt     = (ReaderTest*)calloc(1, sizeof(ReaderTest));
 	SerdReader* reader = serd_reader_new(
 		SERD_TURTLE, rt, free,
 		NULL, NULL, test_sink, NULL);
